@@ -28,6 +28,8 @@ from . import utils as mod_utils
 from . import geo as mod_geo
 from . import gpxfield as mod_gpxfield
 
+log = mod_logging.getLogger(__name__)
+
 # GPX date format to be used when writing the GPX output:
 DATE_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
 
@@ -71,6 +73,7 @@ GPX_10_POINT_FIELDS = [
         mod_gpxfield.GPXField('dgps_id', 'dgpsid'),
 ]
 GPX_11_POINT_FIELDS = [
+        # See GPX for description of text fields
         mod_gpxfield.GPXField('latitude', attribute='lat', type=mod_gpxfield.FLOAT_TYPE, mandatory=True),
         mod_gpxfield.GPXField('longitude', attribute='lon', type=mod_gpxfield.FLOAT_TYPE, mandatory=True),
         mod_gpxfield.GPXField('elevation', 'ele', type=mod_gpxfield.FLOAT_TYPE),
@@ -81,7 +84,7 @@ GPX_11_POINT_FIELDS = [
         mod_gpxfield.GPXField('comment', 'cmt'),
         mod_gpxfield.GPXField('description', 'desc'),
         mod_gpxfield.GPXField('source', 'src'),
-        'link',
+        'link:@link',
             mod_gpxfield.GPXField('link', attribute='href'),
             mod_gpxfield.GPXField('link_text', tag='text'),
             mod_gpxfield.GPXField('link_type', tag='type'),
@@ -95,7 +98,7 @@ GPX_11_POINT_FIELDS = [
         mod_gpxfield.GPXField('position_dilution', 'pdop', type=mod_gpxfield.FLOAT_TYPE),
         mod_gpxfield.GPXField('age_of_dgps_data', 'ageofdgpsdata', type=mod_gpxfield.FLOAT_TYPE),
         mod_gpxfield.GPXField('dgps_id', 'dgpsid'),
-        mod_gpxfield.GPXExtensionsField('extensions'),
+        mod_gpxfield.GPXExtensionsField('extensions', is_list=True),
 ]
 
 # GPX1.0 track points have two more fields after time
@@ -155,9 +158,6 @@ class GPXBounds:
     def __iter__(self):
         return (self.min_latitude, self.max_latitude, self.min_longitude, self.max_longitude,).__iter__()
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
-
 
 class GPXXMLSyntaxException(GPXException):
     """
@@ -210,7 +210,7 @@ class GPXWaypoint(mod_geo.Location):
         self.position_dilution = position_dilution
         self.age_of_dgps_data = None
         self.dgps_id = None
-        self.extensions = None
+        self.extensions = []
 
     def __str__(self):
         return '[wpt{%s}:%s,%s@%s]' % (self.name, self.latitude, self.longitude, self.elevation)
@@ -224,15 +224,29 @@ class GPXWaypoint(mod_geo.Location):
                 representation += ', %s=%s' % (attribute, repr(value))
         return 'GPXWaypoint(%s)' % representation
 
+    def adjust_time(self, delta):
+        """
+        Adjusts the time of the point by the specified delta
+
+        Parameters
+        ----------
+        delta : datetime.timedelta
+            Positive time delta will adjust time into the future
+            Negative time delta will adjust time into the past
+        """
+        if self.time:
+            self.time += delta
+
+    def remove_time(self):
+        """ Will remove time metadata. """
+        self.time = None
+
     def get_max_dilution_of_precision(self):
         """
         Only care about the max dop for filtering, no need to go into too much detail
         """
         return max(self.horizontal_dilution, self.vertical_dilution, self.position_dilution)
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
-
 
 class GPXRoutePoint(mod_geo.Location):
     gpx_10_fields = GPX_10_POINT_FIELDS
@@ -274,7 +288,7 @@ class GPXRoutePoint(mod_geo.Location):
         self.age_of_dgps_data = None
         self.dgps_id = None
         self.link_type = None
-        self.extensions = None
+        self.extensions = []
 
     def __str__(self):
         return '[rtept{%s}:%s,%s@%s]' % (self.name, self.latitude, self.longitude, self.elevation)
@@ -288,8 +302,22 @@ class GPXRoutePoint(mod_geo.Location):
                 representation += ', %s=%s' % (attribute, repr(value))
         return 'GPXRoutePoint(%s)' % representation
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
+    def adjust_time(self, delta):
+        """
+        Adjusts the time of the point by the specified delta
+
+        Parameters
+        ----------
+        delta : datetime.timedelta
+            Positive time delta will adjust time into the future
+            Negative time delta will adjust time into the past
+        """
+        if self.time:
+            self.time += delta
+
+    def remove_time(self):
+        """ Will remove time metadata. """
+        self.time = None
 
 
 class GPXRoute:
@@ -304,18 +332,19 @@ class GPXRoute:
             mod_gpxfield.GPXComplexField('points', tag='rtept', classs=GPXRoutePoint, is_list=True),
     ]
     gpx_11_fields = [
+            # See GPX for description of text fields
             mod_gpxfield.GPXField('name'),
             mod_gpxfield.GPXField('comment', 'cmt'),
             mod_gpxfield.GPXField('description', 'desc'),
             mod_gpxfield.GPXField('source', 'src'),
-            'link',
+            'link:@link',
                 mod_gpxfield.GPXField('link', attribute='href'),
                 mod_gpxfield.GPXField('link_text', tag='text'),
                 mod_gpxfield.GPXField('link_type', tag='type'),
             '/link',
             mod_gpxfield.GPXField('number', type=mod_gpxfield.INT_TYPE),
             mod_gpxfield.GPXField('type'),
-            mod_gpxfield.GPXExtensionsField('extensions'),
+            mod_gpxfield.GPXExtensionsField('extensions', is_list=True),
             mod_gpxfield.GPXComplexField('points', tag='rtept', classs=GPXRoutePoint, is_list=True),
     ]
 
@@ -334,7 +363,25 @@ class GPXRoute:
         self.points = []
         self.link_type = None
         self.type = None
-        self.extensions = None
+        self.extensions = []
+
+    def adjust_time(self, delta):
+        """
+        Adjusts the time of the all the points in the route by the specified delta.
+
+        Parameters
+        ----------
+        delta : datetime.timedelta
+            Positive time delta will adjust time into the future
+            Negative time delta will adjust time into the past
+        """
+        for point in self.points:
+            point.adjust_time(delta)
+
+    def remove_time(self):
+        """ Removes time meta data from route. """
+        for point in self.points:
+            point.remove_time()
 
     def remove_elevation(self):
         """ Removes elevation data from route """
@@ -428,9 +475,6 @@ class GPXRoute:
         for route_point in self.points:
             route_point.move(location_delta)
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
-
     def __repr__(self):
         representation = ''
         for attribute in 'name', 'description', 'number':
@@ -481,7 +525,7 @@ class GPXTrackPoint(mod_geo.Location):
         self.position_dilution = position_dilution
         self.age_of_dgps_data = None
         self.dgps_id = None
-        self.extensions = None
+        self.extensions = []
 
     def __repr__(self):
         representation = '%s, %s' % (self.latitude, self.longitude)
@@ -570,9 +614,6 @@ class GPXTrackPoint(mod_geo.Location):
     def __str__(self):
         return '[trkpt:%s,%s@%s@%s]' % (self.latitude, self.longitude, self.elevation, self.time)
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
-
 
 class GPXTrackSegment:
     gpx_10_fields = [
@@ -580,14 +621,14 @@ class GPXTrackSegment:
     ]
     gpx_11_fields = [
             mod_gpxfield.GPXComplexField('points', tag='trkpt', classs=GPXTrackPoint, is_list=True),
-            mod_gpxfield.GPXExtensionsField('extensions'),
+            mod_gpxfield.GPXExtensionsField('extensions', is_list=True),
     ]
 
     __slots__ = ('points', 'extensions', )
 
     def __init__(self, points=None):
         self.points = points if points else []
-        self.extensions = None
+        self.extensions = []
 
     def simplify(self, max_distance=None):
         """
@@ -912,11 +953,11 @@ class GPXTrackSegment:
 
         if 0 < point_no < len(self.points):
             previous_point = self.points[point_no - 1]
-        if 0 < point_no < len(self.points) - 1:
+        if 0 <= point_no < len(self.points) - 1:
             next_point = self.points[point_no + 1]
 
-        #mod_logging.debug('previous: %s' % previous_point)
-        #mod_logging.debug('next: %s' % next_point)
+        #log.debug('previous: %s' % previous_point)
+        #log.debug('next: %s' % next_point)
 
         speed_1 = point.speed_between(previous_point)
         speed_2 = point.speed_between(next_point)
@@ -943,7 +984,7 @@ class GPXTrackSegment:
         delta : float
             Elevation delta in meters to apply to track
         """
-        mod_logging.debug('delta = %s' % delta)
+        log.debug('delta = %s' % delta)
 
         if not delta:
             return
@@ -1036,11 +1077,11 @@ class GPXTrackSegment:
             last = self.points[-2]
 
         if not last.time or not first.time:
-            mod_logging.debug('Can\'t find time')
+            log.debug('Can\'t find time')
             return None
 
         if last.time < first.time:
-            mod_logging.debug('Not enough time data')
+            log.debug('Not enough time data')
             return None
 
         return mod_utils.total_seconds(last.time - first.time)
@@ -1108,11 +1149,11 @@ class GPXTrackSegment:
         last_time = self.points[-1].time
 
         if not first_time and not last_time:
-            mod_logging.debug('No times for track segment')
+            log.debug('No times for track segment')
             return None
 
         if not first_time <= time <= last_time:
-            mod_logging.debug('Not in track (search for:%s, start:%s, end:%s)' % (time, first_time, last_time))
+            log.debug('Not in track (search for:%s, start:%s, end:%s)' % (time, first_time, last_time))
             return None
 
         for point in self.points:
@@ -1293,8 +1334,6 @@ class GPXTrackSegment:
 
         return len(self.points) > 2 and float(found) / float(len(self.points)) > .75
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
 
     def __repr__(self):
         return 'GPXTrackSegment(points=[%s])' % ('...' if self.points else '')
@@ -1315,18 +1354,19 @@ class GPXTrack:
             mod_gpxfield.GPXComplexField('segments', tag='trkseg', classs=GPXTrackSegment, is_list=True),
     ]
     gpx_11_fields = [
+            # See GPX for text field description
             mod_gpxfield.GPXField('name'),
             mod_gpxfield.GPXField('comment', 'cmt'),
             mod_gpxfield.GPXField('description', 'desc'),
             mod_gpxfield.GPXField('source', 'src'),
-            'link',
+            'link:@link',
                 mod_gpxfield.GPXField('link', attribute='href'),
                 mod_gpxfield.GPXField('link_text', tag='text'),
                 mod_gpxfield.GPXField('link_type', tag='type'),
             '/link',
             mod_gpxfield.GPXField('number', type=mod_gpxfield.INT_TYPE),
             mod_gpxfield.GPXField('type'),
-            mod_gpxfield.GPXExtensionsField('extensions'),
+            mod_gpxfield.GPXExtensionsField('extensions', is_list=True),
             mod_gpxfield.GPXComplexField('segments', tag='trkseg', classs=GPXTrackSegment, is_list=True),
     ]
 
@@ -1345,7 +1385,7 @@ class GPXTrack:
         self.segments = []
         self.link_type = None
         self.type = None
-        self.extensions = None
+        self.extensions = []
 
     def simplify(self, max_distance=None):
         """
@@ -1850,8 +1890,6 @@ class GPXTrack:
     def clone(self):
         return mod_copy.deepcopy(self)
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
 
     def __repr__(self):
         representation = ''
@@ -1880,27 +1918,34 @@ class GPX:
             mod_gpxfield.GPXComplexField('routes', classs=GPXRoute, tag='rte', is_list=True),
             mod_gpxfield.GPXComplexField('tracks', classs=GPXTrack, tag='trk', is_list=True),
     ]
+    # Text fields serialize as empty container tags, dependents are
+    # are listed after as 'tag:dep1:dep2:dep3'. If no dependents are
+    # listed, it will always serialize. The container is closed with
+    # '/tag'. Required dependents are preceded by an @. If a required
+    # dependent is empty, nothing in the container will serialize. The
+    # format is 'tag:@dep2'. No optional dependents need to be listed.
+    # Extensions not yet supported
     gpx_11_fields = [
             mod_gpxfield.GPXField('version', attribute=True),
             mod_gpxfield.GPXField('creator', attribute=True),
-            'metadata',
+            'metadata:name:description:author_name:author_email:author_link:copyright_author:copyright_year:copyright_license:link:time:keywords:bounds',
                 mod_gpxfield.GPXField('name', 'name'),
                 mod_gpxfield.GPXField('description', 'desc'),
-                'author',
+                'author:author_name:author_email:author_link',
                     mod_gpxfield.GPXField('author_name', 'name'),
                     mod_gpxfield.GPXEmailField('author_email', 'email'),
-                    'link',
+                    'link:@author_link',
                         mod_gpxfield.GPXField('author_link', attribute='href'),
                         mod_gpxfield.GPXField('author_link_text', tag='text'),
                         mod_gpxfield.GPXField('author_link_type', tag='type'),
                     '/link',
                 '/author',
-                'copyright',
+                'copyright:copyright_author:copyright_year:copyright_license',
                     mod_gpxfield.GPXField('copyright_author', attribute='author'),
                     mod_gpxfield.GPXField('copyright_year', tag='year'),
                     mod_gpxfield.GPXField('copyright_license', tag='license'),
                 '/copyright',
-                'link',
+                'link:@link',
                     mod_gpxfield.GPXField('link', attribute='href'),
                     mod_gpxfield.GPXField('link_text', tag='text'),
                     mod_gpxfield.GPXField('link_type', tag='type'),
@@ -1913,7 +1958,7 @@ class GPX:
             mod_gpxfield.GPXComplexField('waypoints', classs=GPXWaypoint, tag='wpt', is_list=True),
             mod_gpxfield.GPXComplexField('routes', classs=GPXRoute, tag='rte', is_list=True),
             mod_gpxfield.GPXComplexField('tracks', classs=GPXTrack, tag='trk', is_list=True),
-            mod_gpxfield.GPXExtensionsField('extensions'),
+            mod_gpxfield.GPXExtensionsField('extensions', is_list=True),
     ]
 
     __slots__ = ('version', 'creator', 'name', 'description', 'author_name',
@@ -1921,7 +1966,8 @@ class GPX:
                  'bounds', 'waypoints', 'routes', 'tracks', 'author_link',
                  'author_link_text', 'author_link_type', 'copyright_author',
                  'copyright_year', 'copyright_license', 'link_type',
-                 'metadata_extensions', 'extensions')
+                 'metadata_extensions', 'extensions', 'nsmap',
+                 'schema_locations')
 
     def __init__(self):
         self.version = None
@@ -1942,11 +1988,13 @@ class GPX:
         self.copyright_author = None
         self.copyright_year = None
         self.copyright_license = None
-        self.metadata_extensions = None
-        self.extensions = None
+        self.metadata_extensions = []
+        self.extensions = []
         self.waypoints = []
         self.routes = []
         self.tracks = []
+        self.nsmap = {}
+        self.schema_locations = []
 
     def simplify(self, max_distance=None):
         """
@@ -1990,29 +2038,54 @@ class GPX:
             track.reduce_points(min_distance)
 
         # TODO
-        mod_logging.debug('Track reduced to %s points' % self.get_track_points_no())
+        log.debug('Track reduced to %s points' % self.get_track_points_no())
 
-    def adjust_time(self, delta):
+    def adjust_time(self, delta, all=False):
         """
         Adjusts the time of all points in all of the segments of all tracks by
         the specified delta.
 
+        If all=True, waypoints and routes will also be adjusted by the specified delta.
+
         Parameters
         ----------
         delta : datetime.timedelta
             Positive time delta will adjust times into the future
             Negative time delta will adjust times into the past
+        all : bool
+            When true, also adjusts time for waypoints and routes.
         """
         if self.time:
             self.time += delta
         for track in self.tracks:
             track.adjust_time(delta)
 
-    def remove_time(self):
-        """ Removes time data. """
+        if all:
+            for waypoint in self.waypoints:
+                waypoint.adjust_time(delta)
+            for route in self.routes:
+                route.adjust_time(delta)
+
+    def remove_time(self, all=False):
+        """
+        Removes time data of all points in all of the segments of all tracks.
+
+        If all=True, time date will also be removed from waypoints and routes.
+
+        Parameters
+        ----------
+        all : bool
+            When true, also removes time data for waypoints and routes.
+        """
         for track in self.tracks:
             track.remove_time()
 
+        if all:
+            for waypoint in self.waypoints:
+                waypoint.remove_time()
+            for route in self.routes:
+                route.remove_time()
+
     def remove_elevation(self, tracks=True, routes=False, waypoints=False):
         """ Removes elevation data. """
         if tracks:
@@ -2188,7 +2261,7 @@ class GPX:
     def length_2d(self):
         """
         Computes 2-dimensional length of the GPX file (only latitude and
-        longitude, no elevation). This is the sum of 3D length of all segments
+        longitude, no elevation). This is the sum of 2D length of all segments
         in all tracks.
 
         Returns
@@ -2199,7 +2272,7 @@ class GPX:
         result = 0
         for track in self.tracks:
             length = track.length_2d()
-            if length or length == 0:
+            if length:
                 result += length
         return result
 
@@ -2216,7 +2289,7 @@ class GPX:
         result = 0
         for track in self.tracks:
             length = track.length_3d()
-            if length or length == 0:
+            if length:
                 result += length
         return result
 
@@ -2481,10 +2554,10 @@ class GPX:
 
     def add_missing_elevations(self):
         def _add(interval, start, end, distances_ratios):
+            if (start.elevation is None) or (end.elevation is None):
+                return
             assert start
             assert end
-            assert start.elevation is not None
-            assert end.elevation is not None
             assert interval
             assert len(interval) == len(distances_ratios)
             for i in range(len(interval)):
@@ -2531,7 +2604,7 @@ class GPX:
             time_dist_after = (interval[-1].time_difference(end),
                                interval[-1].distance_3d(end))
 
-            # Assemble list of times and distance to neighboring points
+            # Assemble list of times and distance to neighbour points
             times_dists = [(interval[i].time_difference(interval[i+1]),
                             interval[i].distance_3d(interval[i+1]))
                             for i in range(len(interval) - 1)]
@@ -2546,6 +2619,54 @@ class GPX:
         self.add_missing_data(get_data_function=lambda point: point.speed,
                               add_missing_function=_add)
 
+    def fill_time_data_with_regular_intervals(self, start_time=None, time_delta=None, end_time=None, force=True):
+        """
+        Fills the time data for all points in the GPX file. At least two of the parameters start_time, time_delta, and
+        end_time have to be provided. If the three are provided, time_delta will be ignored and will be recalculated
+        using start_time and end_time.
+
+        The first GPX point will have a time equal to start_time. Then points are assumed to be recorded at regular
+        intervals time_delta.
+
+        If the GPX file currently contains time data, it will be overwritten, unless the force flag is set to False, in
+        which case the function will return a GPXException error.
+
+        Parameters
+        ----------
+        start_time: datetime.datetime object
+            Start time of the GPX file (corresponds to the time of the first point)
+        time_delta: datetime.timedelta object
+            Time interval between two points in the GPX file
+        end_time: datetime.datetime object
+            End time of the GPX file (corresponds to the time of the last point)
+        force: bool
+            Overwrite current data if the GPX file currently contains time data
+        """
+        if not (start_time and end_time) and not (start_time and time_delta) and not (time_delta and end_time):
+            raise GPXException('You must provide at least two parameters among start_time, time_step, and end_time')
+
+        if self.has_times() and not force:
+            raise GPXException('GPX file currently contains time data. Use force=True to overwrite.')
+
+        point_no = self.get_points_no()
+
+        if start_time and end_time:
+            if start_time > end_time:
+                raise GPXException('Invalid parameters: end_time must occur after start_time')
+            time_delta = (end_time - start_time) / (point_no - 1)
+        elif not start_time:
+            start_time = end_time - (point_no - 1) * time_delta
+
+        self.time = start_time
+
+        i = 0
+        for point in self.walk(only_points=True):
+            if i == 0:
+                point.time = start_time
+            else:
+                point.time = start_time + i * time_delta
+            i += 1
+
     def move(self, location_delta):
         """
         Moves each point in the gpx file (routes, waypoints, tracks).
@@ -2564,7 +2685,7 @@ class GPX:
         for track in self.tracks:
             track.move(location_delta)
 
-    def to_xml(self, version=None):
+    def to_xml(self, version=None, prettyprint=True):
         """
         FIXME: Note, this method will change self.version
         """
@@ -2572,7 +2693,7 @@ class GPX:
             if self.version:
                 version = self.version
             else:
-                version = '1.0'
+                version = '1.1'
 
         if version != '1.0' and version != '1.1':
             raise GPXException('Invalid version %s' % version)
@@ -2581,14 +2702,30 @@ class GPX:
         if not self.creator:
             self.creator = 'gpx.py -- https://github.com/tkrajina/gpxpy'
 
-        v = version.replace('.', '/')
-        xml_attributes = {
-                'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
-                'xmlns': 'http://www.topografix.com/GPX/%s' % v,
-                'xsi:schemaLocation': 'http://www.topografix.com/GPX/%s http://www.topografix.com/GPX/%s/gpx.xsd' % (v, v)
-        }
+        self.nsmap['xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
 
-        content = mod_gpxfield.gpx_fields_to_xml(self, 'gpx', version, custom_attributes=xml_attributes)
+        version_path = version.replace('.', '/')
+
+        self.nsmap['defaultns'] = 'http://www.topografix.com/GPX/{0}'.format(
+            version_path
+        )
+
+        if not self.schema_locations:
+            self.schema_locations = [
+                p.format(version_path) for p in (
+                    'http://www.topografix.com/GPX/{0}',
+                    'http://www.topografix.com/GPX/{0}/gpx.xsd',
+                )
+            ]
+
+        content = mod_gpxfield.gpx_fields_to_xml(
+            self, 'gpx', version,
+            custom_attributes={
+                'xsi:schemaLocation': ' '.join(self.schema_locations)
+            },
+            nsmap=self.nsmap,
+            prettyprint=prettyprint
+        )
 
         return '<?xml version="1.0" encoding="UTF-8"?>\n' + content.strip()
 
@@ -2614,9 +2751,6 @@ class GPX:
 
         return result
 
-    def __hash__(self):
-        return mod_utils.hash_object(self, self.__slots__)
-
     def __repr__(self):
         representation = ''
         for attribute in 'waypoints', 'routes', 'tracks':
