@@ -21,26 +21,29 @@ import xml.sax.saxutils as mod_saxutils
 PYTHON_VERSION = mod_sys.version.split(' ')[0]
 
 
-def to_xml(tag, attributes=None, content=None, default=None, escape=False):
+def to_xml(tag, attributes=None, content=None, default=None, escape=False, prettyprint=True, indent=''):
+    if not prettyprint:
+        indent = ''
     attributes = attributes or {}
-    result = '\n<%s' % tag
+    result = []
+    result.append('\n' + indent + '<{0}'.format(tag))
 
     if content is None and default:
         content = default
 
     if attributes:
         for attribute in attributes.keys():
-            result += make_str(' %s="%s"' % (attribute, attributes[attribute]))
+            result.append(make_str(' %s="%s"' % (attribute, attributes[attribute])))
 
     if content is None:
-        result += '/>'
+        result.append('/>')
     else:
         if escape:
-            result += make_str('>%s</%s>' % (mod_saxutils.escape(content), tag))
+            result.append(make_str('>%s</%s>' % (mod_saxutils.escape(content), tag)))
         else:
-            result += make_str('>%s</%s>' % (content, tag))
+            result.append(make_str('>%s</%s>' % (content, tag)))
 
-    result = make_str(result)
+    result = make_str(''.join(result))
 
     return result
 
@@ -74,42 +77,15 @@ def total_seconds(timedelta):
         return None
     return (timedelta.days * 86400) + timedelta.seconds
 
-# Hash utilities:
-
-
-def __hash(obj):
-    result = 0
-
-    if obj is None:
-        return result
-    elif isinstance(obj, dict):
-        raise RuntimeError('__hash_single_object for dict not yet implemented')
-    elif isinstance(obj, list) or isinstance(obj, tuple):
-        return hash_list_or_tuple(obj)
-
-    return hash(obj)
-
-
-def hash_list_or_tuple(iteration):
-    result = 17
-
-    for obj in iteration:
-        result = result * 31 + __hash(obj)
-
-    return result
-
-
-def hash_object(obj, attributes):
-    result = 19
-
-    for attribute in attributes:
-        result = result * 31 + __hash(getattr(obj, attribute))
-
-    return result
-
 
 def make_str(s):
-    """ Convert a str or unicode object into a str type. """
+    """ Convert a str or unicode or float object into a str type. """
+    if isinstance(s, float):
+        result = str(s)
+        if not 'e' in result:
+            return result
+        # scientific notation is illegal in GPX 1/1
+        return format(s, '.10f').rstrip('0.')
     if PYTHON_VERSION[0] == '2':
         if isinstance(s, unicode):
             return s.encode("utf-8")
